from morphocut import Node, Output, ReturnOutputs
import numpy as np
from morphocut.image import RegionProperties, RawOrVariable
from skimage.measure import label as sklabel
from scipy import ndimage as ndi


class DummyRegionProps:
    def __init__(self, shape):
        small = 1e-6

        self.area = 1
        self.perimeter = small
        self.filled_area = 1
        self.min_intensity = 0
        self.max_intensity = 0
        self.mean_intensity = 0
        self.image = np.zeros(shape, dtype=bool)
        self.bbox = (0, 0, 1, 1)
        self.bbox_area = 1
        self.centroid = (0.0, 0.0)
        self.equivalent_diameter = small
        self.major_axis_length = small
        self.minor_axis_length = small
        self.orientation = 0.0
        self.solidity = small
        self.extent = small
        self.eccentricity = small
        self.convex_area = 1
        self.image_filled = np.zeros(shape, dtype=bool)
        self.coords = np.empty((0, 2))
        self.label = -1
        self.euler_number = 0
        self.local_centroid = (0.0, 0.0)


@ReturnOutputs
@Output("regionprops")
class SafeImageProperties(Node):
    def __init__(self, mask: RawOrVariable, image: RawOrVariable = None):
        super().__init__()
        self.mask = mask
        self.image = image

    def transform(self, mask: np.ndarray, image: np.ndarray):
        binmask = (mask != 0)
        if not np.any(binmask):
            print("[WARNING] Empty mask — returning DummyRegionProps.")
            return DummyRegionProps(mask.shape)

        labels, nlabels = sklabel(binmask, return_num=True)

        if nlabels == 1:
            chosen_label = 1
        else:
            counts = np.bincount(labels.ravel())
            counts[0] = 0  # ignore background
            chosen_label = int(np.argmax(counts))

        obj_slices = ndi.find_objects(labels, nlabels)
        slices = obj_slices[chosen_label - 1]
        if slices is None:
            return DummyRegionProps(mask.shape)

        return RegionProperties(
            slices,
            chosen_label,
            labels,
            image,
            True,
        )


import matplotlib.pyplot as plt
from io import BytesIO
from PIL import Image
import numpy as np


def plot_segmentation_result(img_org, initial_contour, refined_contours, title="Segmentation Result"):
    """
    Plot the initial and refined contours on the original image,
    and return the plot as a NumPy array (RGB image).
    """
    fig, ax = plt.subplots(figsize=(8, 8))
    ax.imshow(img_org, cmap='gray')

    if initial_contour is not None:
        ax.plot(initial_contour[:, 1], initial_contour[:, 0], '--r', label='Initial')

    for i, contour in enumerate(refined_contours):
        label = "Refined" if i == 0 else None
        ax.plot(contour[:, 1], contour[:, 0], '-b', linewidth=2, label=label)

    ax.set_title(title)
    ax.legend()
    plt.tight_layout()

    # Save figure to buffer
    buf = BytesIO()
    fig.savefig(buf, format='png')
    plt.close(fig)  # Close the figure to avoid GUI issues

    # Convert buffer to NumPy array
    buf.seek(0)
    image = Image.open(buf).convert('RGB')
    image_np = np.array(image)

    return image_np





import cv2
import numpy as np
from skimage.exposure import match_histograms


def normalize(mask, range_val=255):
    """Normalize an image to a given range."""
    return (range_val * (mask - np.min(mask)) / np.ptp(mask)).astype(np.uint8)


def find_darkest_point(image, mask):
    """Find the darkest pixel within a mask."""
    return np.unravel_index(np.argmin(image[mask]), image.shape)


def apply_histogram_matching(img, ref_path):
    """Match histogram of image to reference."""
    ref = cv2.imread(ref_path, 0)
    matched = match_histograms(img, ref)
    matched = np.clip(matched, img.min(), img.max())
    return matched.astype(img.dtype)


import os
from morphocut.core import Pipeline, Call
from morphocut.file import Find, Glob
from morphocut.image import ImageReader, FindRegions
from holocontour.image.safe_image_properties import SafeImageProperties
from morphocut.stream import Progress
from morphocut.str import Format
from morphocut.contrib.ecotaxa import EcotaxaWriter
from morphocut.contrib.zooprocess import CalculateZooProcessFeatures
from holocontour.contour.morpho_custom import HoloContourNode


def pipeline_run(input_folder, output_name, contour_params, lat=None, lon=None, date=None, ext=".png"):
    output_dir = os.path.join(input_folder, "output")
    os.makedirs(output_dir, exist_ok=True)

    print(f"Selected folder: {input_folder}")
    print(f"Files will be extracted to: {output_dir}")

    with Pipeline() as pipeline:
        fn = Find(input_folder, [ext])
        path = Glob(fn)
        basename = Call(lambda x: os.path.splitext(os.path.basename(x))[0], path)

        metadata = {
            "id": Format("{object_id}", object_id=basename),
            "lat": lat,
            "lon": lon,
            "date": date,
        }

        img = ImageReader(path)

        if img.shape == 2:
            img_gray = img[:, :, 0]
        else:
            img_gray = img

        mask, plot = HoloContourNode(img_gray, contour_params=contour_params)

        region_props = SafeImageProperties(mask, img_gray)

        # to handle multiple objects
        # region_props = FindRegions(mask, img_gray, min_area=30)

        object_meta = CalculateZooProcessFeatures(region_props, metadata)

        EcotaxaWriter(
            os.path.join(output_dir, f"EcoTaxa_{output_name}.zip"),
            (basename + ".jpg", img_gray),
            object_meta,
        )

        EcotaxaWriter(
            os.path.join(output_dir, f"EcoTaxa_{output_name}_mask.zip"),
            (basename + ".jpg", mask),
            object_meta,
        )

        if contour_params.get("save_plot"):
            EcotaxaWriter(
                os.path.join(output_dir, f"EcoTaxa_{output_name}_plot.zip"),
                (basename + ".jpg", plot),
                object_meta,
            )

        Progress(fn)

    pipeline.run()
